# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Declare the UPSTAGE version."""

__authors__ = "UPSTAGE Contributors, GTRI"
__version__ = "0.3.2"
